package com.letmetrial.api.auth.controller;

import java.net.URI;

import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import com.letmetrial.api.auth.provider.AuthenticationProvider;
import com.letmetrial.api.auth.usecase.AtualizarSenhaUC;
import com.letmetrial.api.auth.usecase.AtualizarTokenUC;
import com.letmetrial.api.auth.usecase.CadastrarContaMedicoUC;
import com.letmetrial.api.auth.usecase.LogarUC;

import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import io.swagger.v3.oas.annotations.security.SecurityRequirement;
import jakarta.validation.Valid;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
@RestController
@RequestMapping("auth")
public class AuthController {

    private final AuthenticationProvider authenticationProvider;

    private final CadastrarContaMedicoUC cadastrarContaMedicoUC;

    private final LogarUC logarUC;

    private final AtualizarTokenUC atualizarTokenUC;

    private final AtualizarSenhaUC atualizarSenhaUC;

    @Operation(summary = "Registrar um novo usuário", description = "Registra um novo usuário no sistema", security = @SecurityRequirement(name = "bearerAuth"))
    @ApiResponses(value = {
            @ApiResponse(responseCode = "201", content = @Content(schema = @Schema(implementation = Void.class))),
            @ApiResponse(responseCode = "400", content = @Content(schema = @Schema(implementation = String.class))),
            @ApiResponse(responseCode = "409", content = @Content(schema = @Schema(implementation = String.class))),
    })
    @PostMapping("/register")
    public ResponseEntity<Void> registrarUsuario(@RequestBody @Valid CadastrarContaMedicoUC.NovaContaDTO novaConta) {
        return ResponseEntity
                .created(URI.create(String.format("medicos/%s", cadastrarContaMedicoUC.executar(novaConta).id())))
                .build();
    }

    @Operation(summary = "Logar com usuário registrado", description = "Loga com usuário registrado no sistema")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200", content = @Content(schema = @Schema(implementation = Void.class))),
            @ApiResponse(responseCode = "400", content = @Content(schema = @Schema(implementation = String.class))),
            @ApiResponse(responseCode = "409", content = @Content(schema = @Schema(implementation = String.class))),
    })
    @PostMapping(path = "/login", consumes = { MediaType.APPLICATION_FORM_URLENCODED_VALUE })
    public ResponseEntity<LogarUC.TokenDTO> login(
            @ModelAttribute("email") String email,
            @Schema(format = "password") @ModelAttribute("password") String password) {
        return ResponseEntity.ok(logarUC.executar(email, password));
    }

    @Operation(summary = "Atualizar Token", description = "Atualiza o token de acesso através de um refreshToken")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200", content = @Content(schema = @Schema(implementation = Void.class))),
            @ApiResponse(responseCode = "400", content = @Content(schema = @Schema(implementation = String.class))),
    })
    @PostMapping("/refresh-token")
    public ResponseEntity<AtualizarTokenUC.TokenDTO> atualizarToken(@RequestBody String refreshToken) {
        return ResponseEntity.ok(atualizarTokenUC.run(refreshToken));
    }

    @Operation(summary = "Trocar senha", description = "Troca a senha do usuário", security = @SecurityRequirement(name = "bearerAuth"))
    @PostMapping(path = "/trocar-senha", consumes = { MediaType.APPLICATION_FORM_URLENCODED_VALUE })
    public ResponseEntity<Void> trocarSenha(@RequestBody @Schema(format = "password") String password) {

        atualizarSenhaUC.executar(authenticationProvider.authenticatedUserId(), password);

        return ResponseEntity.ok().build();
    }

}
